#include <stdio.h>
#include <string.h>
#include <stdlib.h>

FILE		*dracbin, *sectordata;

#define FIGHTER 0
#define HORIZONTAL 1
#define VERTICAL 2

typedef struct
{
  int           data_location;
  unsigned char player_x;
  unsigned char player_y;
  unsigned char station_count;
  unsigned char orientation;
  unsigned char station_x[8];
  unsigned char station_y[8];
} SECTOR_DATA;


// sprite data
SECTOR_DATA delta[4];

void help()
{
  printf("\n");
  printf("Usage: dracsec draconian.bin [sectors.txt]\n\n");
  printf("[sectors.txt] is an optional ascii text file with sector data for up\n");
  printf("to 4 sectors.  If this file is not provided the program will output\n");
  printf("the current Delta sectors, as found in draconian.bin, formatted for\n");
  printf("the Draconian source code\n\n");
  printf("If sectors.txt is provided then the Delta sectors in draconian.bin\n");
  printf("will be patched with the sector data provided.\n\n");
  printf("Each sector starts with a line containing the player's ship position\n");
  printf("followed by 1 to 8 lines containing station positions.\n\n");
  printf("Positions are X Y values and can either be in the range of 0-255,\n");
  printf("or in the range of -1 to -8 (X) or -1 to -16 (Y) which denote radar\n");
  printf("pixel positions.\n\n");
  printf("Station positions are preceded by H or V to denote a Horizontal or\n");
  printf("Vertical station\n\n");
  printf("Anything after an * is considered a comment and will be ignored.\n\n");
  printf("Example sectors.txt:\n");
  printf("* sector 1\n");
  printf("128 128 * player starts in middle\n");
  printf("h -1 -1 * horizontal station located at top-left radar position\n");
  printf("V 200 0 * vertical station located near top-right of radar\n");
  printf("H 0 200 * horizontal station located near bottom-left of radar\n");
  printf("v -8 -16 * vertical station located at bottom-right of radar\n");
  printf("* sector 2\n");
  printf("...\n\n");
}

int LoadDeltaSectors()
{
  int size;
  int data;
  int next;
  int table;
  int i,j;
  
  size = 0;
  next = 0;
  table = 0;
  
  while ((data = fgetc(dracbin)) != EOF)
  {
    if (data == 0)
    {
      next = 1;
    }
    else
    {
      if (next > 0)
      {
        if (data == next)
          next++;
        else
          next = 0;
      }
    }
    if (next == 0x30)
    {
      if (table == 0)
        table = size - 4*16 + 1;
      else
      {
        printf("ERROR: Two Delta Tables found.\n");
        return 0;
      }
    }
    
    size++;
    if (size > 32*1024)
      break;
  }
  
  if (size != 32*1024)
  {
    printf("ERROR: ROM not 32K\n");
    return 0;
  }
  
  if (table == 0)
  {
    printf("ERROR: Unable to find Delta Sector table\n");
    return 0;
  }
    
  printf("Delta Sector Table at %x\n", table);
  fseek(dracbin, table, SEEK_SET);
  for(i=0;i<4;i++)
  {
    delta[i].data_location = 0;
    delta[i].player_x = 0;
    delta[i].player_y = 0;
    delta[i].station_count = 0;
    delta[i].orientation = 0;
    for(j=0;j<8;j++)
    {
      delta[i].station_x[j] = 0;
      delta[i].station_y[j] = 0;
    }
    for(j=0;j<4;j++)
    {
      if ((data = fgetc(dracbin)) != EOF)
        delta[i].data_location += (data << (j*8));
    }
    printf("Sector %i at %x\n", i, delta[i].data_location);
    if (delta[i].data_location < 0 || delta[i].data_location > 32*1024)
    {
      printf("ERROR: Delta Sector %i at invalid address\n",i);
      return 0;
    }
  }
  
  // load sector data
  for(i=0;i<4;i++)
  {
    fseek(dracbin, delta[i].data_location, SEEK_SET);
    
    if ((data = fgetc(dracbin)) != EOF)      delta[i].player_x = data;
    if ((data = fgetc(dracbin)) != EOF)      delta[i].player_y = data;
    if ((data = fgetc(dracbin)) != EOF)      delta[i].station_count = data;
    if ((data = fgetc(dracbin)) != EOF)      delta[i].orientation = data;
    if (delta[i].station_count == 0 || delta[i].station_count > 8)
    {
      printf("ERROR: Delta Sector %i has invalid station count of %i\n",i, delta[i].station_count);
      return 0;
    }
    
    for(j=0;j<delta[i].station_count; j++)
    {
      if ((data = fgetc(dracbin)) != EOF)      delta[i].station_x[j] = data;
      if ((data = fgetc(dracbin)) != EOF)      delta[i].station_y[j] = data;
    }
    
    if (data == EOF)
    {
      printf("ERROR: Hit EOF while retrieving data for Delta Sector %i\n",i);
      return 0;
    }
  }
  
  printf("\n\n");
  return 1; // if we made it here then valid sector data was found
}

void PrintOrientation(unsigned char orientation, int station)
{
  if (orientation & (1 << station))     printf("VS ");
  else                                  printf("HS ");
  return;
}

int ConvertRadarPixelX(int x, int object)
{
  // radar pixel covers 64x64 section of screen
  // x level data is divided by 2 to save ROM space
  x = abs(x);
  x--;
  x *= (64/2);
  switch(object)
  {
    case FIGHTER:     x += 14;  break; // width=7 so (64-7)/2/2 = 14
    case HORIZONTAL:  x += 9;   break; // width=28 so (64-28)/2/2 = 9
    case VERTICAL:    x += 8;   break; // width=32 so (64-32)/2/2 = 8
  }
  return x;
}

int ConvertRadarPixelY(int y, int object)
{
  // radar pixel covers 64x64 section of screen
  // y level data is divided by 4 to save ROM space
  
  y = abs(y);
  y--;
  y *= (64/4);
  switch(object)
  {
    case FIGHTER:     y += 6; break; // height=16 so (64-16)/2/4 = 6
    case HORIZONTAL:  y += 0; break; // height=65 so (64-65)/2/4 = 0
    case VERTICAL:    y += 1; break; // height=57 so (64/57)/2/4 = 1
  }
  return y;
}

int LoadSectorData()
{
  int linecount;
  char origline[1024];
  char line[1024];
  int i;
  int sector;
  char value[3][1024];
  int count;
  int x,y;
  int orient;
  
  sector = -1;
  linecount = 0;
  
  while(fgets(origline, sizeof(origline), sectordata) != NULL)
  {
    linecount++;
    strcpy(line, origline);
    // ignore anything after *
    for(i=0;i<sizeof(line);i++)
      if (line[i] == '*')         line[i] = 0;
    
    // break apart data
    count = sscanf(line, "%s %s %s", value[0], value[1], value[2]);
    
    if (count == 1)
    {
      printf("Data incorrect on line %i:\n%s\n", linecount, origline);
      return 0;
    }
    else if (count == 2)
    {
      // expect to find X Y values for ship's starting position.
      count = sscanf(value[0], "%i", &x);
      if (count != 1)
      {
        printf("ERROR: Unexpected X value for Ship's starting position.\n");
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      count = sscanf(value[1], "%i", &y);
      if (count != 1)
      {
        printf("ERROR: Unexpected Y value for Ship's starting position.\n");
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      if (x < 0)  x = ConvertRadarPixelX(x, FIGHTER);
      if (y < 0)  y = ConvertRadarPixelY(y, FIGHTER);
      
      if (x < 0 || x > 255)
      {
        printf("ERROR: X value of %i for Ship's starting position is out of range.\n",x);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      
      if (y < 0 || y > 255)
      {
        printf("ERROR: Y value of %i for Ship's starting position is out of range.\n",y);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      
      sector++;
      if (sector > 3)
      {
        printf("ERROR: Too many Sectors in file.\n");
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      delta[sector].player_x = x;
      delta[sector].player_y = y;
      delta[sector].station_count = 0;
      delta[sector].orientation = 0;
      
      printf("Sector %i, ship location %i, %i\n", sector+1, x, y);
    }
    else if (count == 3)
    {
      // expect to find O X Y values for station's orientation and position
      
      delta[sector].station_count++;
      if (sector < 0)
      {
        printf("ERROR: Ship's starting position needs to preceed station position.\n");
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      if (delta[sector].station_count > 8)
      {
        printf("ERROR: Too many stations for sector %i.\n", sector+1);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      
      if      (value[0][0] == 'v' || value[0][0] == 'V') orient = 1 << (delta[sector].station_count - 1);
      else if (value[0][0] == 'h' || value[0][0] == 'H') orient = 0;
      else
      {
        printf("ERROR: Unexpected orientation of %s for station %i, expected H or V.\n", value[0], delta[sector].station_count);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      
      // expect to find X Y values for ship's starting position.
      count = sscanf(value[1], "%i", &x);
      if (count != 1)
      {
        printf("ERROR: Unexpected X value for Station %i's starting position.\n", delta[sector].station_count);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      count = sscanf(value[2], "%i", &y);
      if (count != 1)
      {
        printf("ERROR: Unexpected Y value for Station %i's starting position.\n", delta[sector].station_count);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      if (x < 0)  x = ConvertRadarPixelX(x, orient ? VERTICAL : HORIZONTAL);
      if (y < 0)  y = ConvertRadarPixelY(y, orient ? VERTICAL : HORIZONTAL);
      
      if (x < 0 || x > 255)
      {
        printf("ERROR: X value of %i for Ship's starting position is out of range.\n",x);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      
      if (y < 0 || y > 255)
      {
        printf("ERROR: Y value of %i for Ship's starting position is out of range.\n",y);
        printf("       Problem occurs on line %i:\n%s\n", linecount, origline);
        return 0;
      }
      
      delta[sector].orientation |= orient;
      delta[sector].station_x[delta[sector].station_count-1] = x;
      delta[sector].station_y[delta[sector].station_count-1] = y;
    }
  }
  if (delta[sector].station_count == 0)
  {
    printf("ERROR: There are no stations in Sector %i.\n", sector+1);
    return 0;
    
  }
  return 1; // if we made it here the sector data was parsed without error
}

void UpdateBin()
{
  int i, j;
  printf("Updating BIN\n");
  
  for(i=0;i<4;i++)
  {
    fseek(dracbin, delta[i].data_location, SEEK_SET);
    fputc(delta[i].player_x,      dracbin);
    fputc(delta[i].player_y,      dracbin);
    fputc(delta[i].station_count, dracbin);
    fputc(delta[i].orientation,   dracbin);
    for(j=0;j<delta[i].station_count;j++)
    {
      fputc(delta[i].station_x[j],      dracbin);
      fputc(delta[i].station_y[j],      dracbin);
    }
  }
}

void OutputCcode()
{
  int i, j;
  
  for(i=0;i<4;i++)
  {
    printf("const unsigned char delta_%i[] =\n", (i+1));
    printf("{\n");
    printf("  %4i, %4i, // Star Fighter X, Y location\n", delta[i].player_x, delta[i].player_y);
    printf("  %4i,       // Station Count\n",delta[i].station_count);
    printf("              //    STATION ORIENTATION\n");
    printf("              //  8  7  6  5    4  3  2  1\n");
    printf("  0x%02x,       // ", delta[i].orientation);
    for(j=7;j>=0;j--)
    {
      PrintOrientation(delta[i].orientation, j);
      if (j == 4)   printf("  ");
    }
    printf("\n");
    for(j=0;j<delta[i].station_count;j++)
    {
      printf("  %4i, %4i%s // ",
             delta[i].station_x[j],
             delta[i].station_y[j],
             (j+1 == delta[i].station_count) ? " " : ",");
      PrintOrientation(delta[i].orientation, j);
      printf("Station %i X, Y location\n",
             j+1);
    }
    printf("};\n\n\n");
  }
}

int main(int argn, char **args)
{
	if (argn != 2 && argn != 3)
	{
    help();
    return 1;
	}
	
	dracbin = fopen(args[1],"rb+"); // b for binary, needed for Windows
	if (dracbin == NULL)
	{
		printf("Unable to open dracbin file '%s'\n\n", args[1]);
		return 1;
	}
  
  if (!LoadDeltaSectors())
  {
    printf("Unable to find Delta Sectors in '%s'\n\n", args[1]);
    fclose(dracbin);
    return 1;
  }
  
  if (argn == 2)
  {
    OutputCcode();
    fclose(dracbin);
    return 0;
  }
  else
  {
    sectordata = fopen(args[2],"r"); // no b or + for sector data
    if (sectordata == NULL)
    {
      printf("Unable to open sector file '%s'\n\n", args[2]);
      fclose(dracbin);
      return 1;
    }
    
    if (LoadSectorData())
      UpdateBin();
  }
	
	fclose(dracbin);
	fclose(sectordata);
  
  return 0;
}
